#pragma once

#include "geometrycentral/surface/simple_polygon_mesh.h"

namespace geometrycentral {
namespace surface {

using PolygonSoupMesh = SimplePolygonMesh;

} // namespace surface
} // namespace geometrycentral
