#pragma once

#include "geometrycentral/surface/manifold_surface_mesh.h"

// TODO guard this with a compile flag
// #pragma error

namespace geometrycentral {
namespace surface {
using HalfedgeMesh = ManifoldSurfaceMesh;
}
} // namespace geometrycentral

